use std::fmt::Display;

use self::add::AddSeries;

use leptos::*;
use leptos_router::*;
use serde::{Deserialize, Serialize};
use uuid::Uuid;

pub mod add;
mod history;
pub mod view;

#[component]
pub fn SeriesList() -> impl IntoView {
    let add_series = expect_context::<Action<AddSeries, Result<(), ServerFnError>>>();
    let delete_series = expect_context::<Action<DeleteSeries, Result<(), ServerFnError>>>();
    let all_series = create_resource(
        move || (add_series.version().get(), delete_series.version().get()),
        |_| async move { get_all().await },
    );
    view! {
        <Suspense fallback=loading>
            { move || all_series.get().map(format_all_series)}
        </Suspense>
        <div>
            <A href="add"
                attr:class="btn btn-primary"
            >
                Add Series
            </A>
        </div>
    }
}

fn format_all_series(
    all_series: Result<Vec<Series>, ServerFnError>,
) -> Result<impl IntoView, ServerFnError> {
    Ok(view! {
        <ErrorBoundary fallback=|errors| {
                view! {
                    <div>
                        {errors.get()
                            .into_iter()
                            .map(|(_, e)| e.to_string())
                            .collect::<Vec<_>>()
                            .join(", ")
                        }
                    </div>
                }
            }
        >
            <ul class="list-group mb-3">
                {
                    all_series.map(|all_series| {
                        Some(all_series.into_iter().map(format_series).collect_view())
                    })
                }
            </ul>
        </ErrorBoundary>
    })
}

fn format_series(series: Series) -> impl IntoView {
    let delete_series = expect_context::<Action<DeleteSeries, Result<(), ServerFnError>>>();
    view! {
        <li class="list-group-item">
            <ActionForm action=delete_series>
                <div class="btn-group" style="float: right;">
                    <A href=format!("view/{}", series.id)
                        attr:class="btn btn-sm btn-secondary"
                        attr:style="float: right"
                    >
                        View
                    </A>
                    <input type="submit"
                        class="btn btn-sm btn-secondary"
                        value="Delete"
                    />
                </div>
                <input type="hidden"
                    name="id"
                    value=series.id.to_string() />
            </ActionForm>
            {series.name}
        </li>
    }
}

pub fn loading() -> impl IntoView {
    view! { <div>Loading...</div> }
}

pub fn format_errors(errors: RwSignal<Errors>) -> impl IntoView {
    view! {
        <div>
            {errors.get()
                .into_iter()
                .map(|(_, e)| e.to_string())
                .collect::<Vec<_>>()
                .join(", ")
            }
        </div>
    }
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub enum Source {
    Disney,
    Hulu,
    Paramount,
    Prime,
    Max,
    Netflix,
    Shudder,
}

impl Display for Source {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Source::Disney => f.write_str("Disney"),
            Source::Hulu => f.write_str("Hulu"),
            Source::Paramount => f.write_str("Paramount"),
            Source::Prime => f.write_str("Prime"),
            Source::Max => f.write_str("Max"),
            Source::Netflix => f.write_str("Netflix"),
            Source::Shudder => f.write_str("Shudder"),
        }
    }
}

#[derive(Debug, Serialize, Deserialize, Clone)]
pub struct Series {
    id: Uuid,
    name: String,
    created: String,
    updated: String,
    source: Source,
    source_url: String,
    imdb: Option<String>,
    wikipedia: Option<String>,
}

#[cfg(feature = "ssr")]
mod ssr {
    use super::{Series, Source};
    use sqlx::{
        postgres::PgRow,
        prelude::*,
        types::chrono::{DateTime, Utc},
    };

    impl<'r> FromRow<'r, PgRow> for Series {
        fn from_row(row: &'r PgRow) -> Result<Self, sqlx::Error> {
            let id = row.try_get("id")?;
            let created: DateTime<Utc> = row.try_get("created")?;
            let created = created.format("%Y-%m-%d").to_string();
            let updated: DateTime<Utc> = row.try_get("updated")?;
            let updated = updated.format("%Y-%m-%d").to_string();
            let name = row.try_get("name")?;
            let source: String = row.try_get("source")?;

            let source_enum: Source =
                source
                    .try_into()
                    .map_err(|e: String| sqlx::Error::ColumnDecode {
                        index: "source".to_string(),
                        source: e.into(),
                    })?;
            let source_url = row.try_get("source_url")?;
            let imdb = row.try_get("imdb")?;
            let wikipedia = row.try_get("wikipedia")?;

            Ok(Series {
                id,
                created,
                updated,
                name,
                source_url,
                imdb,
                wikipedia,
                source: source_enum,
            })
        }
    }

    impl TryFrom<String> for Source {
        type Error = String;

        fn try_from(value: String) -> Result<Self, Self::Error> {
            match value.to_lowercase().as_str() {
                "disney" => Ok(Self::Disney),
                "hulu" => Ok(Self::Hulu),
                "paramount" => Ok(Self::Paramount),
                "prime" => Ok(Self::Prime),
                "max" => Ok(Self::Max),
                "netflix" => Ok(Self::Netflix),
                "shudder" => Ok(Self::Shudder),
                _ => Err(format!("Unrecognized source: {value}")),
            }
        }
    }
}

#[server]
pub async fn get_all() -> Result<Vec<Series>, ServerFnError> {
    use std::sync::OnceLock;

    use sqlx::{Pool, Postgres};

    use crate::db::load_statement;

    static SQL: OnceLock<String> = OnceLock::new();

    let pool: Pool<Postgres> = expect_context();

    let s = SQL.get_or_init(|| load_statement("select_all_series.sql"));
    Ok(sqlx::query_as::<_, Series>(&s)
        .persistent(true)
        .fetch_all(&pool)
        .await?)
}

#[server(DeleteSeries)]
pub async fn delete(id: String) -> Result<(), ServerFnError> {
    use std::sync::OnceLock;

    use sqlx::{Pool, Postgres};

    use crate::db::load_statement;

    static SQL: OnceLock<String> = OnceLock::new();

    let pool: Pool<Postgres> = expect_context();

    let s = SQL.get_or_init(|| load_statement("delete_series.sql"));
    sqlx::query(&s)
        .bind(id.parse::<Uuid>()?)
        .execute(&pool)
        .await?;
    Ok(())
}
