use super::{format_errors, history::ViewingList, Series};
use leptos::*;
use leptos_router::*;
use uuid::Uuid;

#[derive(Params, PartialEq)]
pub struct ViewParams {
    id: Uuid,
}

#[component]
pub fn ViewSeries() -> impl IntoView {
    let ps = use_params::<ViewParams>();
    let id = move || {
        ps.with(|ps| {
            ps.as_ref()
                .map(|ps| ps.id)
                .expect("Failed to parse Id value from query string!")
        })
    };
    let details = create_resource(id, |id| async move { get_series_details(id).await });
    view! {
        <div class="card p-3">
            <div class="card-body">
                <Suspense fallback=loading>
                {details.get().map(format_series_details)}
                </Suspense>
            </div>
        </div>
    }
}

fn loading() -> impl IntoView {
    view! {
        <div class="card-heading mb-3">
            <A href="/"
                attr:class="btn btn-close"
                attr:style="float: right;"
            >
                <span style="display: none;">x</span>
            </A>
        </div>
        <h2><span class="placeholder col-4"></span></h2>
        <div class="btn-group mb-3">
            <button class="btn btn-primary"><span class="placeholder col-4"></span></button>
            <button class="btn btn-primary">IMDB</button>
            <button class="btn btn-primary">Wikipedia</button>
        </div>
    }
}

fn format_series_details(
    details: Result<Series, ServerFnError>,
) -> Result<impl IntoView, ServerFnError> {
    Ok(view! {
        <ErrorBoundary fallback=format_errors>
        {details.map(|details| view! {
            <div class="card-heading mb-3">
                <A href="/"
                    attr:class="btn btn-close"
                    attr:style="float: right;"
                >
                    <span style="display: none;">x</span>
                </A>
                <h2>{details.name}</h2>
                <small class="text-body-secondary">created {details.created}, last updated {details.updated}</small>
            </div>
            <div class="btn-group mb-3">
                <a href={details.source_url}
                    class="btn btn-primary"
                    target="_new"
                >
                {details.source.to_string()}
                </a>
                <OptionalUrlButton label="IMDB" opt_url=details.imdb.clone() />
                <OptionalUrlButton label="Wikipedia" opt_url=details.wikipedia.clone() />
            </div>
            <ViewingList series_id=details.id.clone() />
            })
        }
        </ErrorBoundary>
    })
}

#[component]
fn OptionalUrlButton(label: &'static str, opt_url: Option<String>) -> impl IntoView {
    view! {
        <Suspense fallback=|| ()>
            {opt_url.as_ref().map(|url| view! {
                <a href={url}
                    class="btn btn-primary"
                    target="_new"
                >
                {label}
                </a>
            })}
        </Suspense>
    }
}

#[server(GetSeriesDetails)]
pub async fn get_series_details(id: Uuid) -> Result<Series, ServerFnError> {
    use std::sync::OnceLock;

    use sqlx::{Pool, Postgres};

    use crate::db::load_statement;

    static SQL: OnceLock<String> = OnceLock::new();

    let pool = expect_context::<Pool<Postgres>>();
    let s = SQL.get_or_init(|| load_statement("select_series.sql"));
    Ok(sqlx::query_as::<_, Series>(&s)
        .persistent(true)
        .bind(id)
        .fetch_one(&pool)
        .await?)
}
